/*

                       Inertia Realtime 3D Rendering Engine
            Copyright (c) 1997, Alex Chalfin. All Rights Reserved.
           DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED

*/

#include "clip.h"
#include "pool.h"
#include "xform.h"
#include "shade.h"

/*
  The clipping process used here is based on a 90 degree field of view so
  simplified bounds equations can be used for clipping tests.

  Upon entry, all verticies are in the 90 degree fustrum no matter what
  field of view is defined by the programmer (the CamZFac in the view fixes
  this). The view fustrum is also assumed to be square (nxn) whereas the
  screen is nxm (320x200 for example). The square correction is also performed
  by the code (using CamYFac).

  All "unique" features are performed by earlier parts of the code and are
  undone in the perspective transform.

  For more information on the actual process, see Computer Graphics:
  Principals and Practics
*/


/*
 * Internal arrays specific to the clipping process.
 */

Vertex   *IVerts[12];
Vertex   *OVerts[12];
Vertex2d  IUV[12];
Vertex2d  OUV[12];

/*
 * Internal functions specific to the clipping process.
 */

void FustrumClip(ViewObject *View, Triangle *P, int ShadeIndex, int Mask);
void BottomClip(Vertex **InVert, Vertex **OutVerts, Vertex2d *InUV, Vertex2d *OutUV, ViewObject *View, int *NumVert, int ClipParm);
void RightClip (Vertex **InVert, Vertex **OutVerts, Vertex2d *InUV, Vertex2d *OutUV, ViewObject *View, int *NumVert, int ClipParm);
void LeftClip  (Vertex **InVert, Vertex **OutVerts, Vertex2d *InUV, Vertex2d *OutUV, ViewObject *View, int *NumVert, int ClipParm);
void TopClip   (Vertex **InVert, Vertex **OutVerts, Vertex2d *InUV, Vertex2d *OutUV, ViewObject *View, int *NumVert, int ClipParm);
void zClip     (Vertex **InVert, Vertex **OutVerts, Vertex2d *InUV, Vertex2d *OutUV, ViewObject *View, int *NumVert, int ClipParm);

/*
 * Main Clipping function.
 */
void ClipPolygons(ViewObject *View, VectorObject *Obj)
{
  int   Flag1, Flag2, Flag3;
  int   ClipFlag;
  int   TotalClip;
  int   i;
  Triangle  *P;
  Vertex3d *Av, *Bv, *Cv;

  if (!(Obj->ClipMask & ~CLIP_SHADEMASK)) return;

  for (i = 0, P = Obj->Polygons; i < Obj->NumPoly; i++, P++)
  {
    if (P->Visible == 1)
    {
      ClipFlag  = Obj->ClipMask & CLIP_SHADEMASK;
      TotalClip = 0;

      Av = &P->A->Cam;
      Bv = &P->B->Cam;
      Cv = &P->C->Cam;

      if (Obj->ClipMask & CLIP_Z)
      {
        Flag1 = (Av->z >= View->FustrumMinZ);
        Flag2 = (Bv->z >= View->FustrumMinZ);
        Flag3 = (Cv->z >= View->FustrumMinZ);
        if (Flag1 && Flag2 && Flag3)
        {
        }
        else
        {
          ClipFlag |= CLIP_Z;
          if (!Flag1 && !Flag2 && !Flag3)
            TotalClip = 1;
        }
      }

      if (Obj->ClipMask & CLIP_LEFT)
      {
        Flag1 = (Av->x >= -Av->z);
        Flag2 = (Bv->x >= -Bv->z);
        Flag3 = (Cv->x >= -Cv->z);
        if (Flag1 && Flag2 && Flag3)
        {
        }
        else
        {
          ClipFlag |= CLIP_LEFT;
          if (!Flag1 && !Flag2 && !Flag3)
            TotalClip = 1;
        }
      }

      if (Obj->ClipMask & CLIP_RIGHT)
      {
        Flag1 = (Av->x <= Av->z);
        Flag2 = (Bv->x <= Bv->z);
        Flag3 = (Cv->x <= Cv->z);
        if (Flag1 && Flag2 && Flag3)
        {
        }
        else
        {
          ClipFlag |= CLIP_RIGHT;
          if (!Flag1 && !Flag2 && !Flag3)
            TotalClip = 1;
        }
      }

      /* check for bottom clip */
      if (Obj->ClipMask & CLIP_BOTTOM)
      {
        Flag1 = (Av->y >= -Av->z);
        Flag2 = (Bv->y >= -Bv->z);
        Flag3 = (Cv->y >= -Cv->z);
        if (Flag1 && Flag2 && Flag3)
        {
        }
        else
        {
          ClipFlag |= CLIP_BOTTOM;
          if (!Flag1 && !Flag2 && !Flag3)
            TotalClip = 1;
        }
      }

      if (Obj->ClipMask & CLIP_TOP)
      {
        Flag1 = (Av->y <= Av->z);
        Flag2 = (Bv->y <= Bv->z);
        Flag3 = (Cv->y <= Cv->z);
        if (Flag1 && Flag2 && Flag3)
        {
        }
        else
        {
          ClipFlag |= CLIP_TOP;
          if (!Flag1 && !Flag2 && !Flag3)
            TotalClip = 1;
        }
      }

      if (!TotalClip)
      {
        if (ClipFlag & ~CLIP_SHADEMASK)
          FustrumClip(View, P, Obj->RenderMask, ClipFlag);
      }
      else
      {
        P->Visible = 0;
      }
    }
  }
}

void FustrumClip(ViewObject *View, Triangle *P, int ShadeIndex, int Mask)
{
  int     i;
  int     NumVert;
  int     ClipFlag = 0;
  Triangle   *TempPoly;
  Vertex2d  *InUV;
  Vertex2d  *OutUV;
  Vertex2d  *TempUV;
  Vertex   **InVerts;
  Vertex   **OutVerts;
  Vertex   **TempVert;

  for (i = 0; i < 12; i++)
  {
    IVerts[i] = NULL;
    OVerts[i] = NULL;
  }

  for (i = 0; i < 3; i++)
  {
    IVerts[i]     = GrabVertexFromPool();
    if (IVerts[i] == NULL)
      return;
  }

  IVerts[0]->Cam.x = P->A->Cam.x;
  IVerts[0]->Cam.y = P->A->Cam.y;
  IVerts[0]->Cam.z = P->A->Cam.z;

  IVerts[1]->Cam.x = P->B->Cam.x;
  IVerts[1]->Cam.y = P->B->Cam.y;
  IVerts[1]->Cam.z = P->B->Cam.z;

  IVerts[2]->Cam.x = P->C->Cam.x;
  IVerts[2]->Cam.y = P->C->Cam.y;
  IVerts[2]->Cam.z = P->C->Cam.z;

  if (Mask & CLIP_GOURAUD)
  {
    IVerts[0]->Shade = P->A->Shade;
    IVerts[1]->Shade = P->B->Shade;
    IVerts[2]->Shade = P->C->Shade;
  }

  if (Mask & CLIP_ENV)
  {
    IVerts[0]->Env.x = P->A->Env.x;
    IVerts[0]->Env.y = P->A->Env.y;
    IVerts[1]->Env.x = P->B->Env.x;
    IVerts[1]->Env.y = P->B->Env.y;
    IVerts[2]->Env.x = P->C->Env.x;
    IVerts[2]->Env.y = P->C->Env.y;
  }

  if (Mask & CLIP_TEXTURE)
  {
    InUV  = IUV;
    OutUV = OUV;
    InUV[0].x = P->UV1.x;
    InUV[0].y = P->UV1.y;
    InUV[1].x = P->UV2.x;
    InUV[1].y = P->UV2.y;
    InUV[2].x = P->UV3.x;
    InUV[2].y = P->UV3.y;
  }

  InVerts   = IVerts;
  OutVerts  = OVerts;

  NumVert   = 3;

  /* z-clip if necessairy */
  if (Mask & CLIP_Z)
  {
    ClipFlag = 1;
    zClip(InVerts, OutVerts, InUV, OutUV, View, &NumVert, Mask);
    SWAP(InVerts, OutVerts, TempVert);
    SWAP(InUV, OutUV, TempUV);
  }

  if ((Mask & CLIP_LEFT) && (NumVert >= 3))
  {
    ClipFlag = 1;
    LeftClip(InVerts, OutVerts, InUV, OutUV, View, &NumVert, Mask);
    SWAP(InVerts, OutVerts, TempVert);
    SWAP(InUV, OutUV, TempUV);
  }


  if ((Mask & CLIP_RIGHT) && (NumVert >= 3))
  {
    ClipFlag = 1;
    RightClip(InVerts, OutVerts, InUV, OutUV, View, &NumVert, Mask);
    SWAP(InVerts, OutVerts, TempVert);
    SWAP(InUV, OutUV, TempUV);
  }

  if ((Mask & CLIP_BOTTOM) && (NumVert >= 3))
  {
    ClipFlag = 1;
    BottomClip(InVerts, OutVerts, InUV, OutUV, View, &NumVert, Mask);
    SWAP(InVerts, OutVerts, TempVert);
    SWAP(InUV, OutUV, TempUV);
  }

  if ((Mask & CLIP_TOP) && (NumVert >= 3))
  {
    ClipFlag = 1;
    TopClip(InVerts, OutVerts, InUV, OutUV, View, &NumVert, Mask);
    SWAP(InVerts, OutVerts, TempVert);
    SWAP(InUV, OutUV, TempUV);
  }

  if (ClipFlag)
    P->Visible = 0;

  if (ClipFlag && (NumVert >= 3))
  {
    for (i = 0; i < NumVert; i++)
      SingleProject(View, InVerts[i]);

    for (i = 0; i < (NumVert - 2); i++)
    {
      TempPoly = GrabPolygonFromPool();
      if (TempPoly == NULL)
        return;

      TempPoly->Color = P->Color;
      TempPoly->Shade = P->Shade;
      TempPoly->A     = InVerts[0];
      TempPoly->B     = InVerts[i+1];
      TempPoly->C     = InVerts[i+2];

      if (Mask & CLIP_ENV)
      {
        TempPoly->Texture = P->Texture;
        TempPoly->IBuffer = P->IBuffer;
      }
      if (Mask & CLIP_TEXTURE)
      {
        TempPoly->Texture = P->Texture;
        TempPoly->IBuffer = P->IBuffer;
        TempPoly->UV1.x = InUV[0].x;
        TempPoly->UV1.y = InUV[0].y;
        TempPoly->UV2.x = InUV[i+1].x;
        TempPoly->UV2.y = InUV[i+1].y;
        TempPoly->UV3.x = InUV[i+2].x;
        TempPoly->UV3.y = InUV[i+2].y;
      }

      AddPolygonToRenderList(TempPoly, ShadeIndex, 1);
    }
  }
}

void zClip(Vertex **InVerts, Vertex **OutVerts, Vertex2d *InUV, Vertex2d *OutUV, ViewObject *View, int *NumVert, int ClipParm)
{
  int  OutVertCount = 0;
  int  Curr, CurrFlag;
  int  Next, NextFlag;
  int  VertCount = *NumVert;
  float MinZ      = View->FustrumMinZ;
  float tParm;

  CurrFlag = (InVerts[0]->Cam.z >= MinZ);

  for (Curr = 0; Curr < VertCount; Curr++)
  {
    Next     = (Curr + 1) % VertCount;
    NextFlag = (InVerts[Next]->Cam.z >= MinZ);

    if (CurrFlag)
    {
      if (OutVerts[OutVertCount] == NULL)
      {
        OutVerts[OutVertCount] = GrabVertexFromPool();
        if (OutVerts[OutVertCount] == NULL)
          return;
      }
      OutVerts[OutVertCount]->Cam.x = InVerts[Curr]->Cam.x;
      OutVerts[OutVertCount]->Cam.y = InVerts[Curr]->Cam.y;
      OutVerts[OutVertCount]->Cam.z = InVerts[Curr]->Cam.z;
      OutVerts[OutVertCount]->Shade = InVerts[Curr]->Shade;

      if (ClipParm & CLIP_ENV)
      {
        OutVerts[OutVertCount]->Env.x = InVerts[Curr]->Env.x;
        OutVerts[OutVertCount]->Env.y = InVerts[Curr]->Env.y;
      }
      if (ClipParm & CLIP_TEXTURE)
      {
        OutUV[OutVertCount].x = InUV[Curr].x;
        OutUV[OutVertCount].y = InUV[Curr].y;
      }
      OutVertCount++;
    }

    if (CurrFlag != NextFlag)
    {
      if (OutVerts[OutVertCount] == NULL)
      {
        OutVerts[OutVertCount] = GrabVertexFromPool();
        if (OutVerts[OutVertCount] == NULL)
          return;
      }

      tParm = (MinZ - InVerts[Curr]->Cam.z) / (InVerts[Next]->Cam.z - InVerts[Curr]->Cam.z);

      OutVerts[OutVertCount]->Cam.x = InVerts[Curr]->Cam.x + tParm * (InVerts[Next]->Cam.x - InVerts[Curr]->Cam.x);
      OutVerts[OutVertCount]->Cam.y = InVerts[Curr]->Cam.y + tParm * (InVerts[Next]->Cam.y - InVerts[Curr]->Cam.y);
      OutVerts[OutVertCount]->Cam.z = MinZ;

      if (ClipParm & CLIP_GOURAUD)
      {
        OutVerts[OutVertCount]->Shade = InVerts[Curr]->Shade + tParm * (InVerts[Next]->Shade - InVerts[Curr]->Shade);
      }
      if (ClipParm & CLIP_ENV)
      {
        OutVerts[OutVertCount]->Env.x = InVerts[Curr]->Env.x + tParm * (InVerts[Next]->Env.x - InVerts[Curr]->Env.x);
        OutVerts[OutVertCount]->Env.y = InVerts[Curr]->Env.y + tParm * (InVerts[Next]->Env.y - InVerts[Curr]->Env.y);
      }
      if (ClipParm & CLIP_TEXTURE)
      {
        OutUV[OutVertCount].x = InUV[Curr].x + tParm * (InUV[Next].x - InUV[Curr].x);
        OutUV[OutVertCount].y = InUV[Curr].y + tParm * (InUV[Next].y - InUV[Curr].y);
      }

      OutVertCount++;
    }

    CurrFlag = NextFlag;
  }
  *NumVert = OutVertCount;
}

void LeftClip(Vertex **InVerts, Vertex **OutVerts, Vertex2d *InUV, Vertex2d *OutUV, ViewObject *View, int *NumVert, int ClipParm)
{
  int  OutVertCount;
  int  Curr, CurrFlag;
  int  Next, NextFlag;
  int  VertCount = *NumVert;
  float tParm;
  float dx, dz;

  OutVertCount = 0;

  CurrFlag = (InVerts[0]->Cam.x >= (-InVerts[0]->Cam.z));

  for (Curr = 0; Curr < VertCount; Curr++)
  {
    Next     = (Curr + 1) % VertCount;
    NextFlag = (InVerts[Next]->Cam.x >= (-InVerts[Next]->Cam.z));

    if (CurrFlag)
    {
      if (OutVerts[OutVertCount] == NULL)
      {
        OutVerts[OutVertCount] = GrabVertexFromPool();
        if (OutVerts[OutVertCount] == NULL)
          return;
      }
      OutVerts[OutVertCount]->Cam.x = InVerts[Curr]->Cam.x;
      OutVerts[OutVertCount]->Cam.y = InVerts[Curr]->Cam.y;
      OutVerts[OutVertCount]->Cam.z = InVerts[Curr]->Cam.z;
      OutVerts[OutVertCount]->Shade = InVerts[Curr]->Shade;

      if (ClipParm & CLIP_ENV)
      {
        OutVerts[OutVertCount]->Env.x = InVerts[Curr]->Env.x;
        OutVerts[OutVertCount]->Env.y = InVerts[Curr]->Env.y;
      }
      if (ClipParm & CLIP_TEXTURE)
      {
        OutUV[OutVertCount].x = InUV[Curr].x;
        OutUV[OutVertCount].y = InUV[Curr].y;
      }
      OutVertCount++;
    }

    if (CurrFlag != NextFlag)
    {
      if (OutVerts[OutVertCount] == NULL)
      {
        OutVerts[OutVertCount] = GrabVertexFromPool();
        if (OutVerts[OutVertCount] == NULL)
          return;
      }

      dx = (InVerts[Next]->Cam.x - InVerts[Curr]->Cam.x);
      dz = (InVerts[Next]->Cam.z - InVerts[Curr]->Cam.z);

      tParm = - (InVerts[Curr]->Cam.x + InVerts[Curr]->Cam.z) / (dx + dz);

      OutVerts[OutVertCount]->Cam.x = InVerts[Curr]->Cam.x + tParm * (InVerts[Next]->Cam.x - InVerts[Curr]->Cam.x);
      OutVerts[OutVertCount]->Cam.y = InVerts[Curr]->Cam.y + tParm * (InVerts[Next]->Cam.y - InVerts[Curr]->Cam.y);
      OutVerts[OutVertCount]->Cam.z = -OutVerts[OutVertCount]->Cam.x;

      if (ClipParm & CLIP_GOURAUD)
      {
        OutVerts[OutVertCount]->Shade = InVerts[Curr]->Shade + tParm * (InVerts[Next]->Shade - InVerts[Curr]->Shade);
      }
      if (ClipParm & CLIP_ENV)
      {
        OutVerts[OutVertCount]->Env.x = InVerts[Curr]->Env.x + tParm * (InVerts[Next]->Env.x - InVerts[Curr]->Env.x);
        OutVerts[OutVertCount]->Env.y = InVerts[Curr]->Env.y + tParm * (InVerts[Next]->Env.y - InVerts[Curr]->Env.y);
      }
      if (ClipParm & CLIP_TEXTURE)
      {
        OutUV[OutVertCount].x = InUV[Curr].x + tParm * (InUV[Next].x - InUV[Curr].x);
        OutUV[OutVertCount].y = InUV[Curr].y + tParm * (InUV[Next].y - InUV[Curr].y);
      }
      OutVertCount++;
    }

    CurrFlag = NextFlag;
  }

  *NumVert = OutVertCount;
}

void RightClip(Vertex **InVerts, Vertex **OutVerts, Vertex2d *InUV, Vertex2d *OutUV, ViewObject *View, int *NumVert, int ClipParm)
{
  int  OutVertCount = 0;
  int  Curr, CurrFlag;
  int  Next, NextFlag;
  int  VertCount = *NumVert;
  float tParm;
  float dx, dz;

  CurrFlag = (InVerts[0]->Cam.x <= InVerts[0]->Cam.z);

  for (Curr = 0; Curr < VertCount; Curr++)
  {
    Next     = (Curr + 1) % VertCount;
    NextFlag = (InVerts[Next]->Cam.x <= InVerts[Next]->Cam.z);

    if (CurrFlag)
    {
      if (OutVerts[OutVertCount] == NULL)
      {
        OutVerts[OutVertCount] = GrabVertexFromPool();
        if (OutVerts[OutVertCount] == NULL)
          return;
      }
      OutVerts[OutVertCount]->Cam.x = InVerts[Curr]->Cam.x;
      OutVerts[OutVertCount]->Cam.y = InVerts[Curr]->Cam.y;
      OutVerts[OutVertCount]->Cam.z = InVerts[Curr]->Cam.z;
      OutVerts[OutVertCount]->Shade = InVerts[Curr]->Shade;

      if (ClipParm & CLIP_ENV)
      {
        OutVerts[OutVertCount]->Env.x = InVerts[Curr]->Env.x;
        OutVerts[OutVertCount]->Env.y = InVerts[Curr]->Env.y;
      }
      if (ClipParm & CLIP_TEXTURE)
      {
        OutUV[OutVertCount].x = InUV[Curr].x;
        OutUV[OutVertCount].y = InUV[Curr].y;
      }
      OutVertCount++;
    }

    if (CurrFlag != NextFlag)
    {
      if (OutVerts[OutVertCount] == NULL)
      {
        OutVerts[OutVertCount] = GrabVertexFromPool();
        if (OutVerts[OutVertCount] == NULL)
          return;
      }

      dx = (InVerts[Next]->Cam.x - InVerts[Curr]->Cam.x);
      dz = (InVerts[Next]->Cam.z - InVerts[Curr]->Cam.z);

      tParm = (-InVerts[Curr]->Cam.x + InVerts[Curr]->Cam.z) / (dx - dz);

      OutVerts[OutVertCount]->Cam.x = InVerts[Curr]->Cam.x + tParm * (InVerts[Next]->Cam.x - InVerts[Curr]->Cam.x);
      OutVerts[OutVertCount]->Cam.y = InVerts[Curr]->Cam.y + tParm * (InVerts[Next]->Cam.y - InVerts[Curr]->Cam.y);
      OutVerts[OutVertCount]->Cam.z = OutVerts[OutVertCount]->Cam.x;

      if (ClipParm & CLIP_GOURAUD)
      {
        OutVerts[OutVertCount]->Shade = InVerts[Curr]->Shade + tParm * (InVerts[Next]->Shade - InVerts[Curr]->Shade);
      }
      if (ClipParm & CLIP_ENV)
      {
        OutVerts[OutVertCount]->Env.x = InVerts[Curr]->Env.x + tParm * (InVerts[Next]->Env.x - InVerts[Curr]->Env.x);
        OutVerts[OutVertCount]->Env.y = InVerts[Curr]->Env.y + tParm * (InVerts[Next]->Env.y - InVerts[Curr]->Env.y);
      }
      if (ClipParm & CLIP_TEXTURE)
      {
        OutUV[OutVertCount].x = InUV[Curr].x + tParm * (InUV[Next].x - InUV[Curr].x);
        OutUV[OutVertCount].y = InUV[Curr].y + tParm * (InUV[Next].y - InUV[Curr].y);
      }
      OutVertCount++;
    }

    CurrFlag = NextFlag;
  }

  *NumVert = OutVertCount;
}

void BottomClip(Vertex **InVerts, Vertex **OutVerts, Vertex2d *InUV, Vertex2d *OutUV, ViewObject *View, int *NumVert, int ClipParm)
{
  int  OutVertCount = 0;
  int  Curr, CurrFlag;
  int  Next, NextFlag;
  int  VertCount = *NumVert;
  float tParm;
  float dy, dz;

  CurrFlag = (InVerts[0]->Cam.y >= -InVerts[0]->Cam.z);

  for (Curr = 0; Curr < VertCount; Curr++)
  {
    Next     = (Curr + 1) % VertCount;
    NextFlag = (InVerts[Next]->Cam.y >= -InVerts[Next]->Cam.z);

    if (CurrFlag)
    {
      if (OutVerts[OutVertCount] == NULL)
      {
        OutVerts[OutVertCount] = GrabVertexFromPool();
        if (OutVerts[OutVertCount] == NULL)
          return;
      }
      OutVerts[OutVertCount]->Cam.x = InVerts[Curr]->Cam.x;
      OutVerts[OutVertCount]->Cam.y = InVerts[Curr]->Cam.y;
      OutVerts[OutVertCount]->Cam.z = InVerts[Curr]->Cam.z;
      OutVerts[OutVertCount]->Shade = InVerts[Curr]->Shade;

      if (ClipParm & CLIP_ENV)
      {
        OutVerts[OutVertCount]->Env.x = InVerts[Curr]->Env.x;
        OutVerts[OutVertCount]->Env.y = InVerts[Curr]->Env.y;
      }
      if (ClipParm & CLIP_TEXTURE)
      {
        OutUV[OutVertCount].x = InUV[Curr].x;
        OutUV[OutVertCount].y = InUV[Curr].y;
      }
      OutVertCount++;
    }

    if (CurrFlag != NextFlag)
    {
      if (OutVerts[OutVertCount] == NULL)
      {
        OutVerts[OutVertCount] = GrabVertexFromPool();
        if (OutVerts[OutVertCount] == NULL)
          return;
      }

      dy = (InVerts[Next]->Cam.y - InVerts[Curr]->Cam.y);
      dz = (InVerts[Next]->Cam.z - InVerts[Curr]->Cam.z);

      tParm = -(InVerts[Curr]->Cam.y + InVerts[Curr]->Cam.z) / (dy + dz);

      OutVerts[OutVertCount]->Cam.x = InVerts[Curr]->Cam.x + tParm * (InVerts[Next]->Cam.x - InVerts[Curr]->Cam.x);
      OutVerts[OutVertCount]->Cam.y = InVerts[Curr]->Cam.y + tParm * (InVerts[Next]->Cam.y - InVerts[Curr]->Cam.y);
      OutVerts[OutVertCount]->Cam.z = -OutVerts[OutVertCount]->Cam.y;

      if (ClipParm & CLIP_GOURAUD)
      {
        OutVerts[OutVertCount]->Shade = InVerts[Curr]->Shade + tParm * (InVerts[Next]->Shade - InVerts[Curr]->Shade);
      }

      if (ClipParm & CLIP_ENV)
      {
        OutVerts[OutVertCount]->Env.x = InVerts[Curr]->Env.x + tParm * (InVerts[Next]->Env.x - InVerts[Curr]->Env.x);
        OutVerts[OutVertCount]->Env.y = InVerts[Curr]->Env.y + tParm * (InVerts[Next]->Env.y - InVerts[Curr]->Env.y);
      }
      if (ClipParm & CLIP_TEXTURE)
      {
        OutUV[OutVertCount].x = InUV[Curr].x + tParm * (InUV[Next].x - InUV[Curr].x);
        OutUV[OutVertCount].y = InUV[Curr].y + tParm * (InUV[Next].y - InUV[Curr].y);
      }
      OutVertCount++;
    }

    CurrFlag = NextFlag;
  }

  *NumVert = OutVertCount;
}


void TopClip(Vertex **InVerts, Vertex **OutVerts, Vertex2d *InUV, Vertex2d *OutUV, ViewObject *View, int *NumVert, int ClipParm)
{
  int  OutVertCount = 0;
  int  Curr, CurrFlag;
  int  Next, NextFlag;
  int  VertCount = *NumVert;
  float tParm;
  float dy, dz;

  CurrFlag = (InVerts[0]->Cam.y <= InVerts[0]->Cam.z);

  for (Curr = 0; Curr < VertCount; Curr++)
  {
    Next     = (Curr + 1) % VertCount;
    NextFlag = (InVerts[Next]->Cam.y <= InVerts[Next]->Cam.z);

    if (CurrFlag)
    {
      if (OutVerts[OutVertCount] == NULL)
      {
        OutVerts[OutVertCount] = GrabVertexFromPool();
        if (OutVerts[OutVertCount] == NULL)
          return;
      }
      OutVerts[OutVertCount]->Cam.x = InVerts[Curr]->Cam.x;
      OutVerts[OutVertCount]->Cam.y = InVerts[Curr]->Cam.y;
      OutVerts[OutVertCount]->Cam.z = InVerts[Curr]->Cam.z;
      OutVerts[OutVertCount]->Shade = InVerts[Curr]->Shade;

      if (ClipParm & CLIP_ENV)
      {
        OutVerts[OutVertCount]->Env.x = InVerts[Curr]->Env.x;
        OutVerts[OutVertCount]->Env.y = InVerts[Curr]->Env.y;
      }
      if (ClipParm & CLIP_TEXTURE)
      {
        OutUV[OutVertCount].x = InUV[Curr].x;
        OutUV[OutVertCount].y = InUV[Curr].y;
      }
      OutVertCount++;
    }

    if (CurrFlag != NextFlag)
    {
      if (OutVerts[OutVertCount] == NULL)
      {
        OutVerts[OutVertCount] = GrabVertexFromPool();
        if (OutVerts[OutVertCount] == NULL)
          return;
      }

      dy = (InVerts[Next]->Cam.y - InVerts[Curr]->Cam.y);
      dz = (InVerts[Next]->Cam.z - InVerts[Curr]->Cam.z);

      tParm = (-InVerts[Curr]->Cam.y + InVerts[Curr]->Cam.z) / (dy - dz);

      OutVerts[OutVertCount]->Cam.x = InVerts[Curr]->Cam.x + tParm * (InVerts[Next]->Cam.x - InVerts[Curr]->Cam.x);
      OutVerts[OutVertCount]->Cam.y = InVerts[Curr]->Cam.y + tParm * (InVerts[Next]->Cam.y - InVerts[Curr]->Cam.y);
      OutVerts[OutVertCount]->Cam.z = OutVerts[OutVertCount]->Cam.y;

      if (ClipParm & CLIP_GOURAUD)
      {
        OutVerts[OutVertCount]->Shade = InVerts[Curr]->Shade + tParm * (InVerts[Next]->Shade - InVerts[Curr]->Shade);
      }
      if (ClipParm & CLIP_ENV)
      {
        OutVerts[OutVertCount]->Env.x = InVerts[Curr]->Env.x + tParm * (InVerts[Next]->Env.x - InVerts[Curr]->Env.x);
        OutVerts[OutVertCount]->Env.y = InVerts[Curr]->Env.y + tParm * (InVerts[Next]->Env.y - InVerts[Curr]->Env.y);
      }
      if (ClipParm & CLIP_TEXTURE)
      {
        OutUV[OutVertCount].x = InUV[Curr].x + tParm * (InUV[Next].x - InUV[Curr].x);
        OutUV[OutVertCount].y = InUV[Curr].y + tParm * (InUV[Next].y - InUV[Curr].y);
      }
      OutVertCount++;
    }

    CurrFlag = NextFlag;
  }

  *NumVert = OutVertCount;
}

